/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007-2008, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include "uhub.h"


#ifndef INT_MAX
#define INT_MAX 0x7fffffff
#endif

#ifndef INT_MIN
#define INT_MIN (-0x7fffffff - 1)
#endif

#define CFG_APPLY_BOOLEAN(KEY, TARGET) \
	if (strcmp(KEY, key) == 0) \
	{ \
		if      (strlen(data) == 1 && (data[0] == '1')) TARGET = 1; \
		else if (strlen(data) == 1 && (data[0] == '0')) TARGET = 0; \
		else if (strncasecmp(data, "true",  4) == 0) TARGET = 1; \
		else if (strncasecmp(data, "false", 5) == 0) TARGET = 0; \
		else if (strncasecmp(data, "yes",   3) == 0) TARGET = 1; \
		else if (strncasecmp(data, "no",    2) == 0) TARGET = 0; \
		else if (strncasecmp(data, "on",    2) == 0) TARGET = 1; \
		else if (strncasecmp(data, "off",   3) == 0) TARGET = 0; \
		else\
		{ \
			hub_log(log_fatal, "Configuration error on line %d: '%s' must be either '1' or '0'", line_count, key); \
			return -1; \
		} \
		TARGET |= 0x80000000; \
		return 0; \
	}

#define CFG_APPLY_STRING(KEY, TARGET) \
	if (strcmp(KEY, key) == 0) \
	{ \
		TARGET = hub_strdup(data); \
		return 0; \
	}


#define CFG_APPLY_INTEGER(KEY, TARGET) \
	if (strcmp(KEY, key) == 0) \
	{ \
		char* endptr; \
		int val; \
		errno = 0; \
		val = strtol(data, &endptr, 10); \
		if (((errno == ERANGE && (val == INT_MAX || val == INT_MIN)) || (errno != 0 && val == 0)) || endptr == data /*|| endptr != &data[strlen(data)-1]*/) { \
			hub_log(log_fatal, "Configuration error on line %d: '%s' must be a number", line_count, key); \
			return -1; \
		} \
		TARGET = val; \
		return 0; \
	}


#define DEFAULT_STRING(KEY, VALUE) \
{ \
	if (KEY == 0) \
		KEY = hub_strdup(VALUE); \
}

#define DEFAULT_INTEGER(KEY, VALUE) \
{ \
	if (KEY == 0) \
		KEY = VALUE; \
}

#define DEFAULT_BOOLEAN(KEY, VALUE) \
{ \
	if (KEY & 0x80000000) \
	{ \
		KEY = KEY & 0x000000ff; \
	} \
	else \
	{ \
		KEY = VALUE; \
	} \
}

#define GET_STR(NAME)  CFG_APPLY_STRING ( #NAME , config->NAME )
#define GET_INT(NAME)  CFG_APPLY_INTEGER( #NAME , config->NAME )
#define GET_BOOL(NAME) CFG_APPLY_BOOLEAN( #NAME , config->NAME )
#define IGNORED(NAME) \
    if (strcmp(#NAME, key) == 0) \
    { \
        hub_log(log_warning, "Configuration option %s deprecated and ingnored.", key); \
        return 0; \
    } \


void config_defaults(struct hub_config* config)
{
	DEFAULT_STRING (config->server_bind_addr,      "any");
	DEFAULT_STRING (config->hub_name,              "uhub");
	DEFAULT_STRING (config->hub_description,       "");
	DEFAULT_BOOLEAN(config->hub_enabled,           1);
	DEFAULT_STRING (config->file_acl,              "");
	DEFAULT_STRING (config->file_motd,             "");
	DEFAULT_INTEGER(config->server_port,           1511);
	DEFAULT_INTEGER(config->max_users,             500);
	DEFAULT_INTEGER(config->max_recv_buffer,       4096);
	DEFAULT_INTEGER(config->max_send_buffer,       16384);
	DEFAULT_INTEGER(config->max_send_buffer_soft,  8192);
	DEFAULT_BOOLEAN(config->show_banner,           1);
	DEFAULT_BOOLEAN(config->registered_users_only, 0);
	DEFAULT_BOOLEAN(config->chat_only,             0);
	
	/* Limits enforced on users */
	DEFAULT_INTEGER(config->limit_max_hubs_user,   10);
	DEFAULT_INTEGER(config->limit_max_hubs_reg,    10);
	DEFAULT_INTEGER(config->limit_max_hubs_op,     10);
	DEFAULT_INTEGER(config->limit_min_hubs_user,    0);
	DEFAULT_INTEGER(config->limit_min_hubs_reg,     0);
	DEFAULT_INTEGER(config->limit_min_hubs_op,      0);
	DEFAULT_INTEGER(config->limit_max_hubs,        25);
	DEFAULT_INTEGER(config->limit_min_share,        0);
	DEFAULT_INTEGER(config->limit_max_share,        0);
	DEFAULT_INTEGER(config->limit_min_slots,        1);
	DEFAULT_INTEGER(config->limit_max_slots,        0);

	/* Status/error strings */
	DEFAULT_STRING (config->msg_hub_full,                  "Hub is full");
	DEFAULT_STRING (config->msg_hub_disabled,              "Hub is disabled")
	DEFAULT_STRING (config->msg_hub_registered_users_only, "Hub is for registered users only");
	DEFAULT_STRING (config->msg_inf_error_nick_missing,    "No nickname given");
	DEFAULT_STRING (config->msg_inf_error_nick_multiple,   "Multiple nicknames given");
	DEFAULT_STRING (config->msg_inf_error_nick_invalid,    "Nickname is invalid");
	DEFAULT_STRING (config->msg_inf_error_nick_long,       "Nickname too long");
	DEFAULT_STRING (config->msg_inf_error_nick_short,      "Nickname too short");
	DEFAULT_STRING (config->msg_inf_error_nick_spaces,     "Nickname cannot start with spaces");
	DEFAULT_STRING (config->msg_inf_error_nick_bad_chars,  "Nickname contains invalid characters");
	DEFAULT_STRING (config->msg_inf_error_nick_not_utf8,   "Nickname is not valid utf8");
	DEFAULT_STRING (config->msg_inf_error_nick_taken,      "Nickname is already in use");
	DEFAULT_STRING (config->msg_inf_error_nick_restricted, "Nickname cannot be used on this hub");
	DEFAULT_STRING (config->msg_inf_error_cid_invalid,     "CID is not valid");
	DEFAULT_STRING (config->msg_inf_error_cid_missing,     "CID is not specified");
	DEFAULT_STRING (config->msg_inf_error_cid_taken,       "CID is taken");
	DEFAULT_STRING (config->msg_inf_error_pid_missing,     "PID is not specified");
	DEFAULT_STRING (config->msg_inf_error_pid_invalid,     "PID is invalid");
	DEFAULT_STRING (config->msg_ban_permanently,           "Banned permanently");
	DEFAULT_STRING (config->msg_ban_temporarily,           "Banned temporarily");
	DEFAULT_STRING (config->msg_auth_invalid_password,     "Password is wrong");
	DEFAULT_STRING (config->msg_auth_user_not_found,       "User not found in password database");
	DEFAULT_STRING (config->msg_error_no_memory,           "No memory");

	DEFAULT_INTEGER(config->tls_enable,             0);
	DEFAULT_INTEGER(config->tls_require,            0);
	DEFAULT_STRING (config->tls_certificate,       "");
	DEFAULT_STRING (config->tls_private_key,       "");
}


static int apply_config(struct hub_config* config, char* key, char* data, int line_count)
{

	GET_STR (file_acl);
	GET_STR (file_motd);
	GET_STR (server_bind_addr);
	GET_INT (server_port);
	GET_STR (hub_name);
	GET_STR (hub_description);
	GET_BOOL(hub_enabled);
	GET_INT (max_users);
	GET_INT (max_recv_buffer);
	GET_INT (max_send_buffer);
	GET_INT (max_send_buffer_soft);
	GET_BOOL(show_banner);
	GET_BOOL(chat_only);
	GET_BOOL(registered_users_only);

	/* Limits enforced on users */
	GET_INT(limit_max_hubs_user);
	GET_INT(limit_max_hubs_reg);
	GET_INT(limit_max_hubs_op);
	GET_INT(limit_min_hubs_user);
	GET_INT(limit_min_hubs_reg);
	GET_INT(limit_min_hubs_op);
	GET_INT(limit_max_hubs);
	GET_INT(limit_min_share);
	GET_INT(limit_max_share);
	GET_INT(limit_min_slots);
	GET_INT(limit_max_slots);
	
	/* Status/error strings */
	GET_STR (msg_hub_full);
	GET_STR (msg_hub_disabled);
	GET_STR (msg_hub_registered_users_only);
	GET_STR (msg_inf_error_nick_missing);
	GET_STR (msg_inf_error_nick_multiple);
	GET_STR (msg_inf_error_nick_invalid);
	GET_STR (msg_inf_error_nick_long);
	GET_STR (msg_inf_error_nick_short);
	GET_STR (msg_inf_error_nick_spaces);
	GET_STR (msg_inf_error_nick_bad_chars);
	GET_STR (msg_inf_error_nick_not_utf8);
	GET_STR (msg_inf_error_nick_taken);
	GET_STR (msg_inf_error_nick_restricted);
	GET_STR (msg_inf_error_cid_invalid);
	GET_STR (msg_inf_error_cid_missing);
	GET_STR (msg_inf_error_cid_taken);
	GET_STR (msg_inf_error_pid_missing);
	GET_STR (msg_inf_error_pid_invalid);
	GET_STR (msg_ban_permanently);
	GET_STR (msg_ban_temporarily);
	GET_STR (msg_auth_invalid_password);
	GET_STR (msg_auth_user_not_found);
	GET_STR (msg_error_no_memory);

	GET_BOOL(tls_enable);
	GET_BOOL(tls_require);
	GET_STR (tls_certificate);
	GET_STR (tls_private_key);

    /* Deprectated */
    IGNORED (server_max_fds);
	
	/* Still here -- unknown directive */
	hub_log(log_fatal, "Unknown configuration directive: '%s'", key);
	return -1;
}


void free_config(struct hub_config* config)
{
	hub_free(config->server_bind_addr);
	hub_free(config->file_motd);
	hub_free(config->file_acl);
	hub_free(config->hub_name);
	hub_free(config->hub_description);
	
	hub_free(config->msg_hub_full);
	hub_free(config->msg_hub_disabled);
	hub_free(config->msg_hub_registered_users_only);
	hub_free(config->msg_inf_error_nick_missing);
	hub_free(config->msg_inf_error_nick_multiple);
	hub_free(config->msg_inf_error_nick_invalid);
	hub_free(config->msg_inf_error_nick_long);
	hub_free(config->msg_inf_error_nick_short);
	hub_free(config->msg_inf_error_nick_spaces);
	hub_free(config->msg_inf_error_nick_bad_chars);
	hub_free(config->msg_inf_error_nick_not_utf8);
	hub_free(config->msg_inf_error_nick_taken);
	hub_free(config->msg_inf_error_nick_restricted);
	hub_free(config->msg_inf_error_cid_invalid);
	hub_free(config->msg_inf_error_cid_missing);
	hub_free(config->msg_inf_error_cid_taken);
	hub_free(config->msg_inf_error_pid_missing);
	hub_free(config->msg_inf_error_pid_invalid);
	hub_free(config->msg_ban_permanently);
	hub_free(config->msg_ban_temporarily);
	hub_free(config->msg_auth_invalid_password);
	hub_free(config->msg_auth_user_not_found);
	hub_free(config->msg_error_no_memory);
	
	hub_free(config->tls_certificate);
	hub_free(config->tls_private_key);
	
	memset(config, 0, sizeof(struct hub_config));
}


static int config_parse_line(char* line, int line_count, void* ptr_data)
{
	char* pos;
	char* key;
	char* data;
	struct hub_config* config = (struct hub_config*) ptr_data;

	if ((pos = strchr(line, '#')) != NULL)
	{
		pos[0] = 0;
	}

	if (strlen(line) == 0) return 0;

#ifdef CONFIG_DUMP
	hub_log(log_trace, "config_parse_line(): '%s'", line);
#endif

	if ((pos = strchr(line, '=')) != NULL)
	{
		pos[0] = 0;
	}

	key = line;
	data = &pos[1];

	key = strip_white_space(key);
	data = strip_white_space(data);

	if (!strlen(key) || !strlen(data))
	{
		hub_log(log_fatal, "Configuration parse error on line %d", line_count);
		return -1;
	}

#ifdef CONFIG_DUMP
	hub_log(log_trace, "config_parse_line: '%s' => '%s'", key, data);
#endif

	return apply_config(config, key, data, line_count);
}


int read_config(const char* file, struct hub_config* config)
{
	int ret;

	memset(config, 0, sizeof(struct hub_config));

	ret = file_read_lines(file, config, &config_parse_line);
	if (ret == -1)
		return -1;

	config_defaults(config);
	return 0;
}



